# Copyright (c) 2012, 2022 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================

"""idd comments have gaps in them.
With \note fields as indicated
This code fills those gaps
see: SCHEDULE:DAY:LIST as an example"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

# see iddgaps6.py on usage
# TODO : need unit tests for all htese functions


# works for all objects, including troublesome ones
# factor the code i  iddgaps4.py into functions


# idd might look like this
# <snip>
# / field varA 1
# / field varB 1
# / field varC 1
# / field varA 2
# / field varB 2
# / field varC 2
# / above pattern continues
#
# find objects where the fields are not named
# do the following only for those objects
# find the first field that has an integer. This is a repeating field
# gather the repeating field names (without the integer)
# generate all the repeating fields for all variables


import eppy.bunchhelpers as bunchhelpers


def cleaniddfield(acomm):
    """make all the keys lower case"""
    for key in list(acomm.keys()):
        val = acomm[key]
        acomm[key.lower()] = val
    for key in list(acomm.keys()):
        val = acomm[key]
        if key != key.lower():
            acomm.pop(key)
    return acomm


def cleancommdct(commdct):
    """make all keys in commdct lower case"""
    return [[cleaniddfield(fcomm) for fcomm in comm] for comm in commdct]


def getfields(comm):
    """get all the fields that have the key 'field'"""
    fields = []
    for field in comm:
        if "field" in field:
            fields.append(field)
    return fields


def repeatingfieldsnames(fields, int_replace=None):
    """get the names of the repeating fields"""
    if not int_replace:
        int_replace = "%s"
    fnames = [field["field"][0] for field in fields]
    fnames = [bunchhelpers.onlylegalchar(fname) for fname in fnames]
    fnames = [fname for fname in fnames if bunchhelpers.intinlist(fname.split())]
    fnames = [(bunchhelpers.replaceint(fname, int_replace), None) for fname in fnames]
    dct = dict(fnames)
    repnames = fnames[: len(list(dct.keys()))]
    return repnames


def a_missingkey_standard(commdct, key_i, key_txt, nofirstfields):
    """
    sometimes IDD has fields with no field description. This happens in extensible.
    This functions picks up the previous field description and adds them
    to the fields that have no field descriptions. This update is done in place to commdct
    commdct is not returned.
    missing field descriptions are actually mising keys of a dict
    hence the function is called a missingkey

    This is very old function. Not sure how it works now.
    """
    comm = commdct[key_i]

    # get all fields
    fields = getfields(comm)

    # get repeating field names
    repnames = repeatingfieldsnames(fields)

    try:
        first = repnames[0][0] % (1,)
    except IndexError:
        nofirstfields.append(key_txt)
        return nofirstfields

    # get all comments of the first repeating field names
    firstnames = [repname[0] % (1,) for repname in repnames]
    fcomments = [
        field
        for field in fields
        if bunchhelpers.onlylegalchar(field["field"][0]) in firstnames
    ]
    fcomments = [dict(fcomment) for fcomment in fcomments]
    for cmt in fcomments:
        fld = cmt["field"][0]
        fld = bunchhelpers.onlylegalchar(fld)
        fld = bunchhelpers.replaceint(fld)
        cmt["field"] = [fld]

    for i, cmt in enumerate(comm[1:]):
        thefield = cmt["field"][0]
        thefield = bunchhelpers.onlylegalchar(thefield)
        if thefield == first:
            break
    first_i = i + 1

    newfields = []
    for i in range(1, len(comm[first_i:]) // len(repnames) + 1):
        for fcomment in fcomments:
            nfcomment = dict(fcomment)
            fld = nfcomment["field"][0]
            fld = fld % (i,)
            nfcomment["field"] = [fld]
            newfields.append(nfcomment)

    for i, cmt in enumerate(comm):
        if i < first_i:
            continue
        else:
            afield = newfields.pop(0)
            comm[i] = afield
    commdct[key_i] = comm
    return nofirstfields


def missingkeys_standard(commdct, dtls, skiplist=None):
    """put missing keys in commdct for standard objects
    return a list of keys where it is unable to do so
    commdct is not returned, but is updated"""
    if skiplist == None:
        skiplist = []
    # find objects where all the fields are not named
    gkeys = [dtls[i] for i in range(len(dtls)) if commdct[i].count({}) > 2]
    nofirstfields = []
    # operatie on those fields
    for key_txt in gkeys:
        if key_txt in skiplist:
            continue
            # return nofirstfields
        key_i = dtls.index(key_txt.upper())
        # a_missingkey_standard_1(commdct[key_i], key_txt, nofirstfields)
        a_missingkey_standard(commdct, key_i, key_txt, nofirstfields)
    return nofirstfields


def missingkeys_nonstandard(block, commdct, dtls, objectlist, afield="afiled %s"):
    """This is an object list where thre is no first field name
    to give a hint of what the first field name should be"""
    afield = "afield %s"
    for key_txt in objectlist:
        key_i = dtls.index(key_txt.upper())
        comm = commdct[key_i]
        if block:
            blk = block[key_i]
        for i, cmt in enumerate(comm):
            if cmt == {}:
                first_i = i
                break
        for i, cmt in enumerate(comm):
            if i >= first_i:
                if block:
                    comm[i]["field"] = ["%s" % (blk[i])]
                else:
                    comm[i]["field"] = [afield % (i - first_i + 1,)]



# Copyright (c) 2012 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================

"""pytest for idfreader. very few tests"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from io import StringIO

import eppy.idfreader as idfreader
from eppy.EPlusInterfaceFunctions import readidf

from eppy.iddcurrent import iddcurrent

iddfhandle = StringIO(iddcurrent.iddtxt)


def test_iddversiontuple():
    """py.test for iddversiontuple"""
    iddtxt = """stuff 9.8.4
    other stuff"""
    fhandle = StringIO(iddtxt)
    result = idfreader.iddversiontuple(fhandle)
    assert result == (9, 8, 4)


def test_convertafield():
    """py.test for convertafield"""
    data = (
        ({"type": "integer"}, "1", "N1", 1),
        # field_comm, field_val, field_iddname, expected
        ({}, "1", "N1", 1.0),
        # field_comm, field_val, field_iddname, expected
        ({"type": "real"}, "1", "N1", 1.0),
        # field_comm, field_val, field_iddname, expected
        ({}, "autosize", "N1", "autosize"),
        ({}, "4", "A1", "4"),
        # field_comm, field_val, field_iddname, expected
    )
    for field_comm, field_val, field_iddname, expected in data:
        result = idfreader.convertafield(field_comm, field_val, field_iddname)
        assert result == expected


def test_convertallfields():
    """py.test convertallfields"""
    data = (
        ("version, 8.1;", "VERSION", ["version", "8.1"]),
        # idfstr, objkey, expected
        (
            "WINDOWMATERIAL:SIMPLEGLAZINGSYSTEM, simple, 0.45;",
            "WINDOWMATERIAL:SIMPLEGLAZINGSYSTEM",
            ["WINDOWMATERIAL:SIMPLEGLAZINGSYSTEM", "simple", 0.45],
        ),
        # idfstr, objkey, expected
        (
            "HVACTEMPLATE:ZONE:FANCOIL, gumby1, gumby2, 0.45;",
            "HVACTEMPLATE:ZONE:FANCOIL",
            ["HVACTEMPLATE:ZONE:FANCOIL", "gumby1", "gumby2", 0.45],
        ),
        # idfstr, objkey, expected
        (
            "HVACTEMPLATE:ZONE:FANCOIL, gumby1, gumby2, autosize;",
            "HVACTEMPLATE:ZONE:FANCOIL",
            ["HVACTEMPLATE:ZONE:FANCOIL", "gumby1", "gumby2", "autosize"],
        ),
        # idfstr, objkey, expected
    )
    commdct = None
    block = None
    for idfstr, objkey, expected in data:
        idfhandle = StringIO(idfstr)
        block, data, commdct, idd_index = readidf.readdatacommdct1(
            idfhandle, iddfile=iddfhandle, commdct=commdct, block=block
        )
        idfreader.convertallfields(data, commdct, block)
        result = data.dt[objkey][0]
        assert result == expected



# Copyright (c) 2022 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""py.test to test extending extensible fields beyond what is in the IDD file"""

from io import StringIO

from eppy.iddcurrent import iddcurrent
from eppy.modeleditor import IDF


# idd is read only once in this test
# if it has already been read from some other test, it will continue with
# the old reading
iddfhandle = StringIO(iddcurrent.iddtxt)
if IDF.getiddname() == None:
    IDF.setiddname(iddfhandle)


def test_read_overextended():
    """py.test when reading an IDD that has more fields than the IDD"""
    astr = """
    Version,
        9.6;                      !- Version Identifier

    WINDOWMATERIAL:GLAZINGGROUP:THERMOCHROMIC,
        Gumby,                    !- Name
        55,
        G1,"""

    # nn = 5000
    nn = 50
    extfields = ",".join([f"{i}, G{i}" for i in range(nn)])
    newstr = f"{astr} {extfields};"

    fhandle = StringIO(newstr)
    idf = IDF(fhandle)
    wm = idf.idfobjects["WINDOWMATERIAL:GLAZINGGROUP:THERMOCHROMIC"]
    assert wm[0][f"Optical_Data_Temperature_{nn + 1}"] == nn - 1
    assert wm[0][f"Window_Material_Glazing_Name_{nn + 1}"] == f"G{nn - 1}"


def test_newidfobject_overextend():
    """py.test when idf.newidfobject creates an object with more fields than avaliable in the IDD"""
    idf = IDF(StringIO(""))
    n = 2000
    d1 = dict(Name="Gumby")
    d2 = {f"Optical_Data_Temperature_{i}": i for i in range(1, n + 1)}
    d3 = {f"Window_Material_Glazing_Name_{i}": f"G{i}" for i in range(1, n + 1)}
    kwargs = dict()
    kwargs.update(d1)
    kwargs.update(d2)
    kwargs.update(d3)
    d4 = {"Window_Material_Glazing_Name_2010": "G2010"}  # skip some numbers
    kwargs.update(d4)
    wm = idf.newidfobject("WindowMaterial:GlazingGroup:Thermochromic", **kwargs)
    assert wm.Optical_Data_Temperature_2000 == 2000
    assert wm.Window_Material_Glazing_Name_2000 == "G2000"
    assert wm.Window_Material_Glazing_Name_2010 == "G2010"  # test after skiping
    assert wm.Window_Material_Glazing_Name_2005 == ""  # test the skipped fields


def test_getset_overextended():
    """pytest when idfobject.fieldname and idfobject['fieldname'] is an overextended field"""
    idf = IDF(StringIO(""))
    wm = idf.newidfobject("WindowMaterial:GlazingGroup:Thermochromic", Name="Gumby")
    wm.Optical_Data_Temperature_2000 = 2000  # test __setattr__
    assert wm.Optical_Data_Temperature_2000 == 2000
    wm["Optical_Data_Temperature_2001"] = 2001  # rest __setitem__
    assert wm.Optical_Data_Temperature_2001 == 2001
    assert wm.Optical_Data_Temperature_2002 == ""  # test __getattr__
    assert wm["Optical_Data_Temperature_2003"] == ""  # test __getitem__



# Copyright (c) 2022 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""functions to deal with extensible fields
Moving the functions here is a work in progress"""


import eppy.iddgaps as iddgaps
import eppy.bunchhelpers as bunchhelpers


def getextensible(objidd):
    """return the extensible from the idd"""
    keys = objidd[0].keys()
    extkey = [key for key in keys if key.startswith("extensible")]
    if extkey:
        extens = extkey[0].split(":")[-1]
        return int(extens)
    else:
        return None


def extension_of_extensible(objidd, objblock, n):
    """generate the list of new vars needed to extend by n"""
    ext = getextensible(objidd)
    n = n // ext
    lastvars = endof_extensible(ext, objblock)
    alpha_lastvars = [i[0] for i in lastvars]
    int_lastvars = [int(i[1:]) for i in lastvars]
    lst = []
    for alpha, start in zip(alpha_lastvars, int_lastvars):
        step = alpha_lastvars.count(alpha)
        rng = range(start + step, start + 1 + n * step, step)
        lst.append(["{}{}".format(alpha, item) for item in rng])

    from itertools import chain

    return list(chain(*zip(*lst)))


def endof_extensible(extensible, thisblock):
    """get the vars from where extension happens"""
    return thisblock[-extensible:]


def increaseIDDfields(block, commdct, key_i, key_txt, n):
    """increase fields in IDD - ie in block and commdct"""
    extlst = extension_of_extensible(commdct[key_i], block[key_i], n)
    block[key_i] = block[key_i] + extlst
    commdct[key_i] = commdct[key_i] + [{}] * len(extlst)
    nofirstfields = []
    # print(commdct[key_i])
    iddgaps.a_missingkey_standard(commdct, key_i, key_txt, nofirstfields)
    objfields = [comm.get("field") for comm in commdct[key_i]]
    return objfields


# def increaseIDDfields_1(o_block, o_commdct, key_txt, n):
#     """increase fields in IDD - ie in block and commdct"""
#     print(o_commdct)
#     extlst = extension_of_extensible(o_commdct, o_block, n)
#     # o_block = o_block + extlst
#     o_block.append(extlst)
#     # o_commdct = o_commdct + [{}] * len(extlst)
#     o_commdct.append([{}] * len(extlst))
#     nofirstfields = []
#     # print(o_commdct)
#     iddgaps.a_missingkey_standard_1(o_commdct, key_txt, nofirstfields)
#     objfields = [comm.get("field") for comm in o_commdct]
#     return objfields


def islegalextensiblefield(objidd, fieldname):
    """return true if afieldname is an extensible field in objidd"""
    comm = objidd
    if getextensible(comm):
        if " " not in fieldname:  # a space instead of a underscore will return True
            # the code below is a little wierd
            fields = iddgaps.getfields(comm)
            repnames_tuples = iddgaps.repeatingfieldsnames(fields, "0")
            repnames = [bunchhelpers.makefieldname(item[0]) for item in repnames_tuples]
            fieldname_spaces = fieldname.replace("_", " ")
            fieldname_spaces0 = bunchhelpers.replaceint(fieldname_spaces, "0")
            fieldname0 = bunchhelpers.makefieldname(fieldname_spaces0)
            expected = fieldname0 in repnames
        else:
            expected = False
    return expected


def extfieldint(fieldname, sep=None):
    """return the integer in the extensible field
    It there is more than one integer, it will return the last one"""
    if not sep:
        sep = "_"
    words = fieldname.split(sep)
    digits = [int(word) for word in words if word.isdigit()]
    if digits:
        return digits[-1]
    else:
        return None



# Copyright (c) 2022 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================

"""py.tests for ext_field_functions.py"""

import pytest
import eppy.ext_field_functions as extff


def test_getextensible():
    """py.test for getextensible"""
    data = (
        (
            [
                {
                    "format": ["singleLine"],
                    "group": "Simulation Parameters",
                    "idfobj": "Version",
                    "memo": ["Specifies the EnergyPlus version of the IDF file."],
                    "unique-object": [""],
                },
                {},
                {},
                {},
            ],
            None,
        ),  # objidd, expected
        (
            [
                {
                    "extensible:2": [
                        '- repeat last two fields, remembering to remove ; from "inner" fields.'
                    ],
                    "group": "Schedules",
                    "idfobj": "Schedule:Day:Interval",
                    "memo": [
                        "A Schedule:Day:Interval contains a full day of values with specified end times for each value",
                        "Currently, is set up to allow for 10 minute intervals for an entire day.",
                    ],
                    "min-fields": ["5"],
                },
                {},
                {},
                {},
            ],
            2,
        ),  # objidd, expected
    )
    for objidd, expected in data:
        result = extff.getextensible(objidd)
        assert result == expected


def test_extension_of_extensible():
    """py.test for extension_of_extensible"""
    data = (
        (
            [
                {
                    "extensible:1": None,
                    "group": "Schedules",
                    "idfobj": "Schedule:Day:Interval",
                },
                {},
                {},
                {},
            ],
            ["Gumby", "A1", "A2"],
            1,
            ["A3"],
        ),  # objidd, objblock, n, expected
        (
            [
                {
                    "extensible:1": None,
                    "group": "Schedules",
                    "idfobj": "Schedule:Day:Interval",
                },
                {},
                {},
                {},
            ],
            ["Gumby", "A1", "A2"],
            2,
            ["A3", "A4"],
        ),  # objidd, objblock, n, expected
        (
            [
                {
                    "extensible:1": None,
                    "group": "Schedules",
                    "idfobj": "Schedule:Day:Interval",
                },
                {},
                {},
                {},
            ],
            ["Gumby", "A1", "A2"],
            3,
            ["A3", "A4", "A5"],
        ),  # objidd, objblock, n, expected
        (
            [
                {
                    "extensible:2": None,
                    "group": "Schedules",
                    "idfobj": "Schedule:Day:Interval",
                },
                {},
                {},
                {},
            ],
            ["Gumby", "A1", "A2"],
            2,
            ["A3", "A4"],
        ),  # objidd, objblock, n, expected
        (
            [
                {
                    "extensible:4": None,
                    "group": "Schedules",
                    "idfobj": "Schedule:Day:Interval",
                },
                {},
                {},
                {},
            ],
            ["Gumby", "N3", "A4", "M8", "A5"],
            4,
            ["N4", "A6", "M9", "A7"],
        ),  # objidd, objblock, n, expected
        (
            [
                {
                    "extensible:4": None,
                    "group": "Schedules",
                    "idfobj": "Schedule:Day:Interval",
                },
                {},
                {},
                {},
            ],
            ["Gumby", "N3", "A4", "M8", "A5"],
            12,
            [
                "N4",
                "A6",
                "M9",
                "A7",
                "N5",
                "A8",
                "M10",
                "A9",
                "N6",
                "A10",
                "M11",
                "A11",
            ],
        ),  # objidd, objblock, n, expected
    )
    for objidd, objblock, n, expected in data:
        result = extff.extension_of_extensible(objidd, objblock, n)
        assert result == expected


def test_endof_extensible():
    """py.test for endof_extensible"""
    data = (
        (1, ["gumby", "A1", "A2"], ["A2"]),  # extensible, thisblock, expected
        (1, ["gumby", "A1", "A2", "A3"], ["A3"]),  # extensible, thisblock, expected
        (
            2,
            ["gumby", "A1", "A2", "A3"],
            ["A2", "A3"],
        ),  # extensible, thisblock, expected
    )
    for extensible, thisblock, expected in data:
        result = extff.endof_extensible(extensible, thisblock)
        assert result == expected


@pytest.mark.parametrize(
    "objidd, afieldname, expected",
    [
        (
            [
                {
                    "extensible:2": [
                        '- repeat last two fields remembering to remove ; from "inner" fields.'
                    ]
                },
                {"field": ["Name"]},
                {"field": ["Optical Data Temperature 1"], "begin-extensible": [""]},
                {"field": ["Window Material Glazing Name 1"]},
                {"field": ["Optical Data Temperature 2"]},
                {"field": ["Window Material Glazing Name 2"]},
            ],
            "Optical_Data_Temperature_32",
            True,
        ),  # objidd, afieldname, expected
        (
            [
                {
                    "extensible:2": [
                        '- repeat last two fields remembering to remove ; from "inner" fields.'
                    ]
                },
                {"field": ["Name"]},
                {"field": ["Optical Data Temperature 1"], "begin-extensible": [""]},
                {"field": ["Window Material Glazing Name 1"]},
                {"field": ["Optical Data Temperature 2"]},
                {"field": ["Window Material Glazing Name 2"]},
            ],
            "Gumby_800",
            False,
        ),  # objidd, afieldname, expected
        (
            [
                {
                    "extensible:2": [
                        '- repeat last two fields remembering to remove ; from "inner" fields.'
                    ]
                },
                {"field": ["Name"]},
                {"field": ["Optical Data Temperature 1"], "begin-extensible": [""]},
                {"field": ["Window Material Glazing Name 1"]},
                {"field": ["Optical Data Temperature 2"]},
                {"field": ["Window Material Glazing Name 2"]},
            ],
            "Optical_Data Temperature 32",
            False,
        ),  # objidd, afieldname, expected
    ],
)
def test_islegalextensiblefield(objidd, afieldname, expected):
    """py.test for islegalextensiblefield"""
    result = extff.islegalextensiblefield(objidd, afieldname)
    assert result == expected


@pytest.mark.parametrize(
    "fname, expected",
    [
        ("this_name_4", 4),  # fname, expected
        ("this_3_name_4", 4),  # fname, expected
        ("this_name", None),  # fname, expected
    ],
)
def test_extfieldint(fname, expected):
    """py.test for extfieldint"""
    result = extff.extfieldint(fname)
    assert result == expected
